package se.paxport.yggejonssoncrazysnake;

import java.util.Comparator;

import se.citerus.crazysnake.Direction;

/**
 * @author Joel Jonsson
 */
public class GoScore implements Comparable<GoScore> {
    private MyPos goalPos;

    private Direction gotoDir;

    private int myOpenSpaceLimit;

    private int killOpponentTurn = Integer.MAX_VALUE;

    private boolean eatsFruits = false;

    private boolean stealsFruit = false;

    private int distance = 0;

    private boolean iAmFirst = false;

    private int myOpenSpace;

    private int nearbyFruits;

    private int nearbyFruitScore;

    private static final int REALLY_NEAR_FRUIT_DIST = 7;

    private static final Comparator<GoScore> COMPARATOR = new Comparator<GoScore>() {
        public int compare(GoScore gs1, GoScore gs2) {
            return gs1.compareTo(gs2);
        }
    };

    @Override
    public int compareTo(GoScore o) {
        int cmp;
        if (criticalLowMyOpenSpace() || o.criticalLowMyOpenSpace()) {
            cmp = -Integer.valueOf(myOpenSpace).compareTo(o.myOpenSpace);
            if (cmp != 0) {
                return cmp;
            }
        }

        cmp = Integer.valueOf(killOpponentTurn).compareTo(o.killOpponentTurn);
        if (cmp != 0) {
            return cmp;
        }

        cmp = -Boolean.valueOf(stealsFruit).compareTo(o.stealsFruit);
        if (cmp != 0) {
            return cmp;
        }

        cmp = -Boolean.valueOf(eatsFruits).compareTo(o.eatsFruits);
        if (cmp != 0) {
            return cmp;
        }

        cmp = -Boolean.valueOf(iAmFirst).compareTo(o.iAmFirst);
        if (cmp != 0) {
            return cmp;
        }

        if (myOpenSpace > 1000 && o.myOpenSpace > 1000) {
            if (Math.abs(myOpenSpace - o.myOpenSpace) > 200) {
                cmp = -Integer.valueOf(myOpenSpace).compareTo(o.myOpenSpace);
                if (cmp != 0) {
                    return cmp;
                }
            }
        } else {
            cmp = -Integer.valueOf(myOpenSpace).compareTo(o.myOpenSpace);
            if (cmp != 0) {
                return cmp;
            }
        }
        if (distance <= REALLY_NEAR_FRUIT_DIST || o.distance <= REALLY_NEAR_FRUIT_DIST) {
            cmp = Integer.valueOf(distance).compareTo(o.distance);
            if (cmp != 0) {
                return cmp;
            }
        }
        cmp = Integer.valueOf(distance - nearbyFruitScore).compareTo(o.distance - o.nearbyFruitScore);
        if (cmp != 0) {
            return cmp;
        }

        return 0;
    }

    private boolean criticalLowMyOpenSpace() {
        return myOpenSpace < myOpenSpaceLimit;
    }

    public GoScore(MyPos goalPos, int myOpenSpaceLimit) {
        this.goalPos = goalPos;
        this.myOpenSpaceLimit = myOpenSpaceLimit;
    }

    public void addDistance() {
        distance++;
    }

    public void setGotoDir(Direction gotoDir) {
        this.gotoDir = gotoDir;
    }

    public void eatFruit() {
        eatsFruits = true;
    }

    public void setIAmFirst() {
        iAmFirst = true;
    }

    public void setKillOpponent(int killOpponentTurn) {
        this.killOpponentTurn = killOpponentTurn;
    }


    public boolean isEatingFruit() {
        return eatsFruits;
    }

    public void stealsFruit() {
        stealsFruit = true;
    }

    public void setMyOpenSpace(int myOpenSpace) {
        this.myOpenSpace = myOpenSpace;
    }

    public Direction getDirection() {
        return gotoDir;
    }

    public void addNearbyFruitScore(int fruitScore) {
        nearbyFruitScore += fruitScore;
        nearbyFruits++;
    }

    @Override
    public String toString() {
        return "GoScore{" +
                "goalPos=" + goalPos +
                ", gotoDir=" + gotoDir +
                ", myOpenSpaceLimit=" + myOpenSpaceLimit +
                ", killOpponentTurn=" + killOpponentTurn +
                ", eatsFruits=" + eatsFruits +
                ", stealsFruit=" + stealsFruit +
                ", distance=" + distance +
                ", iAmFirst=" + iAmFirst +
                ", myOpenSpace=" + myOpenSpace +
                ", nearbyFruits=" + nearbyFruits +
                ", nearbyFruitScore=" + nearbyFruitScore +
                '}';
    }
}
