package se.paxport.yggejonssoncrazysnake;

import se.citerus.crazysnake.Direction;
import se.citerus.crazysnake.Position;

/**
 * A position with x and y coordinate on the board.
 */
@SuppressWarnings({"RedundantIfStatement"})
public class MyPos {
    final int x;
    final int y;
    private static MyPos[] positionCache;
    private static int maxX;
    private static int maxY;

    static {
        initCache(50, 50);
    }

    public static void initCache(int maxX, int maxY) {
        if (MyPos.maxX != maxX || MyPos.maxY != maxY) {
            MyPos.maxX = maxX;
            MyPos.maxY = maxY;
            positionCache = new MyPos[maxY * maxX];
            for (int y = 0; y < maxY; y++) {
                for (int x = 0; x < maxX; x++) {
                    positionCache[y * maxX + x] = new MyPos(x, y);
                }
            }
        }
    }

    public static MyPos positionOf(Position p) {
        return positionOf(p.getX(), p.getY());
    }

    public static MyPos positionOf(int x, int y) {
        if (x < 0 || y < 0 || x >= maxX || y >= maxY) return null;
        return positionCache[y * maxX + x];
    }

    private MyPos(int x, int y) {
        this.x = x;
        this.y = y;
    }

    /**
     * Returns the {@code Position} of a square offset with specified length
     * with respect to the specified {@code Direction}.
     *
     * @param direction the offset {@code Direction}.
     * @return the {@code Position} of a square offset with specified length
     *         with respect to the specified {@code Direction}.
     */
    public MyPos offset(Direction direction) {
        switch (direction) {
            case EAST:
                return positionOf(x + 1, y);
            case NORTH:
                return positionOf(x, y - 1);
            case SOUTH:
                return positionOf(x, y + 1);
            case WEST:
                return positionOf(x - 1, y);
            default:
                throw new IllegalArgumentException();
        }
    }

    @Override
    public String toString() {
        return "(" + x + "," + y + ")";
    }

    @Override
    public int hashCode() {
        final int prime = 51;
        int result = 1;
        result = prime * result + x;
        result = prime * result + y;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        MyPos other = (MyPos) obj;
        if (x != other.x)
            return false;
        if (y != other.y)
            return false;
        return true;
    }

    public Position getPosition() {
        return new Position(x, y);
    }
}
