package se.paxport.yggejonssoncrazysnake;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import se.citerus.crazysnake.BaseBrain;
import se.citerus.crazysnake.BrainId;
import se.citerus.crazysnake.Direction;
import se.citerus.crazysnake.HeatMeta;
import se.citerus.crazysnake.HeatState;
import se.citerus.crazysnake.Movement;
import se.citerus.crazysnake.Position;
import se.citerus.crazysnake.Snake;
import se.citerus.crazysnake.SquareContent;

@SuppressWarnings({"UnusedDeclaration"})
public class NewPaxport2 extends BaseBrain {

    public static final int LETHAL = -(1 << 28);

    private static final int FRUIT = -(1 << 27);

    private HeatState state;

    private int maxX;

    private int maxY;

    private MyPos myHeadPosition;

    private MyPos oppHeadPosition;

    @SuppressWarnings({"FieldCanBeLocal"})
    private Direction currentDirection;

    private Direction[] dirValues = Direction.values();

    private Movement[] movValues = Movement.values();

    @SuppressWarnings({"FieldCanBeLocal"})
    private Snake me;

    @SuppressWarnings({"FieldCanBeLocal"})
    private Snake opponent;

    private Map<BrainId, SnakeWithSegments2> holders;

    private SnakeWithSegments2 myHolder;

    private SnakeWithSegments2 oppHolder;

    private int[][] occupied = null;

    /**
     * Gå inte direkt mot den här, använd {@link #getZeroMatrix()} istället.
     */
    private int[][] occupiedTmp;

    private int[] zeroRow;

    private Direction[][] from;

    private int opponentCurrentOpenSpace;

    private int myOpenSpaceLimit;

    private boolean underDog;

    private CircularMyPosQueue q;

    private long timeLimit;

    private int[][] interestingOnTurn;

    private int turn;

    private int manhattan(MyPos p1, MyPos p2) {
        return Math.abs(p1.x - p2.x) + Math.abs(p1.y - p2.y);
    }


    public void init(Set<BrainId> participants, HeatMeta meta) {
        super.init(participants, meta);
        holders = null;


        maxY = (int) Math.round(meta.getBoardDimensions().getHeight());
        maxX = (int) Math.round(meta.getBoardDimensions().getWidth());
        MyPos.initCache(maxX, maxY);

        occupied = new int[maxY][maxX];
        occupiedTmp = new int[maxY][maxX];
        from = new Direction[maxY][maxX];
        interestingOnTurn = new int[maxY][maxX];
        zeroRow = new int[maxX];
        q = new CircularMyPosQueue(maxX * maxY);
        timeLimit = meta.getThinkTimeInMs() * 1000000L * 21 / 25;
    }

    @SuppressWarnings({"ConstantConditions"})
    @Override
    public Movement getNextMove(HeatState state) {
        long startTime = System.nanoTime();
        this.state = state;

        me = getMe(state);
        opponent = getOpponent(state);
        myHeadPosition = MyPos.positionOf(me.getHeadPosition());
        oppHeadPosition = MyPos.positionOf(opponent.getHeadPosition());
        currentDirection = me.getDirection();
        underDog = me.getScore() < opponent.getScore();
        turn = (int) state.getTurn();

        updateHolders(state);
        precalcOccupied();
        myOpenSpaceLimit = Math.min(me.getLength() + (me.getLength() / meta.getGrowFrequency()) + 17, 400);

        opponentCurrentOpenSpace = calculateOpenSpace(oppHeadPosition, 0, opponent.getLength());

        List<MyPos> interestingPositions = getInterestingPositionsAndFillFrom(state);
        GoScore bestGoScore = null;
        for (MyPos interestingPos : interestingPositions) {
            if (System.nanoTime() - startTime > timeLimit) {
                break;
            }
            if (interestingOnTurn[interestingPos.y][interestingPos.x] != turn) {
                continue;
            }
            interestingOnTurn[interestingPos.y][interestingPos.x] = -1;
            GoScore goScore = tryGoal(interestingPos);
            if (goScore == null) {
                continue;
            }
            if (bestGoScore == null || goScore.compareTo(bestGoScore) < 0) {
                bestGoScore = goScore;
            }
        }
        if (bestGoScore == null) {
            return Movement.FORWARD;
        }
        Movement mov = Movement.FORWARD;
        for (Movement movValue : movValues) {
            if (currentDirection.turn(movValue).equals(bestGoScore.getDirection())) {
                mov = movValue;
            }
        }
        return mov;
    }

    private GoScore tryGoal(MyPos goalPos) {
        GoScore goScore = new GoScore(goalPos, myOpenSpaceLimit);
        int len = 0;
        MyPos p = goalPos;
        List<MyPos> vag = new ArrayList<MyPos>();
        Direction goneDir = null;
        while (!p.equals(myHeadPosition)) {
            if (occupied[goalPos.y][goalPos.x] == FRUIT) {
                goScore.eatFruit();
            }
            vag.add(p);
            goScore.addDistance();
            goneDir = from[p.y][p.x];
            p = p.offset(negatDir(goneDir));
        }
        goScore.setGotoDir(goneDir);

        Collections.reverse(vag);

        int[][] g = getZeroMatrix();
        g[oppHeadPosition.y][oppHeadPosition.x] = 1;
        q.clear();
        q.add(oppHeadPosition);

        List<Integer> tmpOnExtraVag = new ArrayList<Integer>();
        int stepsGone = 0;
        boolean iAmFirst = true;
        int oppOpenSpace = 0;
        int oppClosestFruit = Integer.MAX_VALUE;
        List<MyPos> closeOppFruit = new ArrayList<MyPos>();
        for (MyPos myGoPos : vag) {
            if (!underDog) {
                myHolder.populateSnakeShrinkingWithExtraVag(occupied, meta, state, myGoPos);
            }
            stepsGone++;

            while (!q.isEmpty() && g[q.peek().y][q.peek().x] == stepsGone) {
                oppOpenSpace++;
                p = q.remove();

                if (occupied[p.y][p.x] == FRUIT && oppClosestFruit == Integer.MAX_VALUE) {
                    oppClosestFruit = g[p.y][p.x];
                }
                if (oppClosestFruit == g[p.y][p.x]) {
                    closeOppFruit.add(p);
                }


                for (Direction dir : dirValues) {
                    MyPos p2 = p.offset(dir);
                    if (canGo(g, p, p2)) {
                        q.add(p2);
                        g[p2.y][p2.x] = g[p.y][p.x] + 1;
                    }
                }
            }
            if (g[myGoPos.y][myGoPos.x] != 0) {
                iAmFirst = false;
            }
            if (underDog) {
                myHolder.populateSnakeShrinkingWithExtraVag(occupied, meta, state, myGoPos);
            }
        }

        if (iAmFirst) {
            goScore.setIAmFirst();

            while (!q.isEmpty()) {
                oppOpenSpace++;
                p = q.remove();

                if (occupied[p.y][p.x] == FRUIT && oppClosestFruit == Integer.MAX_VALUE) {
                    oppClosestFruit = g[p.y][p.x];
                }
                if (oppClosestFruit == g[p.y][p.x]) {
                    closeOppFruit.add(p);
                }

                for (Direction dir : dirValues) {
                    MyPos p2 = p.offset(dir);
                    if (canGo(g, p, p2)) {
                        q.add(p2);
                        g[p2.y][p2.x] = g[p.y][p.x] + 1;
                        oppOpenSpace++;
                    }
                }
            }
            if (oppOpenSpace < opponent.getLength() && oppOpenSpace < opponentCurrentOpenSpace) {
                goScore.setKillOpponent(oppOpenSpace);
            }

            if (goScore.isEatingFruit() && closeOppFruit.contains(goalPos) && closeOppFruit.size() == 1) {
                goScore.stealsFruit();
            }
        } else {

        }

        goScore.setMyOpenSpace(calculateOpenSpace(goalPos, vag.size(), Integer.MAX_VALUE));
        calculateNearbyFruits(goalPos, vag.size(), goScore);

        myHolder.clearExtraVag(occupied, vag);
        return goScore;
    }

    private boolean canGo(int[][] g, MyPos from, MyPos to) {
        if (g[from.y][from.x] > 1001) {
            return false;
        }
        return to != null
                && occupied[to.y][to.x] != LETHAL
                && g[to.y][to.x] == 0
                && (occupied[to.y][to.x] == FRUIT
                || g[from.y][from.x] >= occupied[to.y][to.x]);
    }

    List<MyPos> getInterestingPositionsAndFillFrom(HeatState state) {
        List<MyPos> interestingPositions = new ArrayList<MyPos>();
        for (Position position : state.getFruitPositions()) {
            interestingPositions.add(MyPos.positionOf(position));
        }

        int[][] g = getZeroMatrix();

        MyPos startPos = myHeadPosition;
        Direction startDir = me.getDirection();

        g[startPos.y][startPos.x] = 1;
        q.clear();
        q.add(startPos);

        while (!q.isEmpty()) {
            MyPos p = q.remove();
            for (Direction dir : dirValues) {
                if (p.equals(startPos) && negatDir(dir).equals(startDir)) {
                    continue;
                }
                MyPos p2 = p.offset(dir);
                if (canGo(g, p, p2)) {
                    g[p2.y][p2.x] = g[p.y][p.x] + 1;
                    if (occupied[p2.y][p2.x] != FRUIT) {
                        interestingPositions.add(p2);
                    }
                    interestingOnTurn[p2.y][p2.x] = turn;
                    from[p2.y][p2.x] = dir;
                    q.add(p2);
                }
            }
        }

        return interestingPositions;
    }

    private void updateHolders(HeatState state) {
        if (holders == null) {
            holders = new HashMap<BrainId, SnakeWithSegments2>();
            for (Snake snake : state.snakes()) {
                holders.put(snake.getId(), new SnakeWithSegments2(snake));
            }
            myHolder = holders.get(me.getId());
            oppHolder = holders.get(opponent.getId());
        }
        myHolder.refreshState(state, myHeadPosition);
        oppHolder.refreshState(state, oppHeadPosition);
    }

    /**
     * Endast en användare av resultatet åt gången tack!
     *
     * @return nollad återanvänd matris
     */
    private int[][] getZeroMatrix() {
        for (int y = 0; y < maxY; y++) {
            System.arraycopy(zeroRow, 0, occupiedTmp[y], 0, maxX);
        }
        return occupiedTmp;
    }

    private void precalcOccupied() {
        for (int y = 0; y < maxY; y++) {
            for (int x = 0; x < maxX; x++) {
                if (isLethalCollision(x, y)) {
                    occupied[y][x] = LETHAL;
                } else if (state.getSquare(x, y).containsFruit()) {
                    occupied[y][x] = FRUIT;
                } else {
                    occupied[y][x] = 0;
                }
            }
        }

        for (SnakeWithSegments2 snakeWithSegments : holders.values()) {
            snakeWithSegments.populateSnakeShrinking(occupied, meta, state);
        }
    }

    private int calculateOpenSpace(MyPos newHead, int turnsPlayedBeforeCalc, int maxLimit) {
        int[][] g = getZeroMatrix();
        g[newHead.y][newHead.x] = 1 + turnsPlayedBeforeCalc;
        q.clear();
        q.add(newHead);

        int openSpace = 0;
        while (!q.isEmpty()) {
            openSpace++;
            if (openSpace > maxLimit) {
                return openSpace;
            }
            MyPos p = q.remove();
            for (Direction dir : dirValues) {
                MyPos p2 = p.offset(dir);
                if (canGo(g, p, p2)) {
                    q.add(p2);
                    g[p2.y][p2.x] = g[p.y][p.x] + 1;
                }
            }
        }
        return openSpace;
    }

    private void calculateNearbyFruits(MyPos newHead, int turnsPlayedBeforeCalc, GoScore goScore) {
        int[][] g = getZeroMatrix();
        int startTurn = 1 + turnsPlayedBeforeCalc;
        g[newHead.y][newHead.x] = startTurn;
        q.clear();
        q.add(newHead);

        while (!q.isEmpty()) {
            MyPos p = q.remove();
            for (Direction dir : dirValues) {
                MyPos p2 = p.offset(dir);
                if (canGo(g, p, p2)) {
                    g[p2.y][p2.x] = g[p.y][p.x] + 1;
                    if (occupied[p2.y][p2.x] == FRUIT) {
                        goScore.addNearbyFruitScore(startTurn + 21 - g[p2.y][p2.x]);
                    }
                    if (g[p2.y][p2.x] <= startTurn + 19) {
                        q.add(p2);
                    }
                }
            }
        }
    }

    private static Direction negatDir(Direction dir) {
        switch (dir) {
            case SOUTH:
                return Direction.NORTH;
            case NORTH:
                return Direction.SOUTH;
            case WEST:
                return Direction.EAST;
            case EAST:
                return Direction.WEST;
        }
        return null;
    }

    private boolean isLethalCollision(int x, int y) {
        if (x < 0 || x >= maxX || y < 0 || y >= maxY) {
            return true;
        }
        SquareContent content = state.getSquare(x, y).getContent();
        return content.getContentType().isSolid();

    }
}
