package se.paxport.yggejonssoncrazysnake;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import se.citerus.crazysnake.BrainId;
import se.citerus.crazysnake.HeatMeta;
import se.citerus.crazysnake.HeatState;
import se.citerus.crazysnake.Snake;
import se.citerus.crazysnake.Square;
import se.citerus.crazysnake.SquareContentType;

/**
 * @author Joel Jonsson
 */
public class SnakeWithSegments2 {

    private final LinkedList<MyPos> segments = new LinkedList<MyPos>();

    private final BrainId id;

    private List<Integer> tmpOnExtraVag = new ArrayList<Integer>();

    private int turnsUntilGrowth;

    private int turnsUntilGone;

    private int turnsUntilGrowthOnTurn;

    private int turnsUntilGoneOnTurn;

    public SnakeWithSegments2(Snake snake) {
        this.id = snake.getId();
        segments.add(MyPos.positionOf(snake.getHeadPosition()));
        if (snake.getLength() != 1) {
            System.out.println(String.format("Snake %s has unexpected length: %d", snake.getName(), snake.getLength()));
        }
    }

    public void refreshState(HeatState state, MyPos headPosition) {
        shrinkTail(state);
        growHead(headPosition);
        if (segments.size() != state.getSnake(id).getLength()) {
            System.out.println("SnakeWithSegments.refreshState length wrong");
            System.out.println("segments.size() = " + segments.size());
            System.out.println("state.getSnake(id).getLength() = " + state.getSnake(id).getLength());
        }
    }

    private void growHead(MyPos headPosition) {
        if (segments.isEmpty() || !segments.getFirst().equals(headPosition)) {
            segments.addFirst(headPosition);
        }
    }

    private void shrinkTail(HeatState heatState) {
        while (!segments.isEmpty()) {
            MyPos tail = segments.getLast();
            Square square = heatState.getSquare(tail.getPosition());
            if (!square.containsSnake()
                    || square.getContent().getContentType() == SquareContentType.SNAKE_HEAD
                    || !id.equals(square.getContent().getSnake().getId())) {
                segments.removeLast();
            } else {
                break;
            }
        }
    }

    public void populateSnakeShrinking(int[][] g, HeatMeta meta, HeatState state) {
        boolean earlyTurn = state.getTurn() <= 10;
        turnsUntilGrowth = (int) state.getTurnsUntilGrowth();
        turnsUntilGone = 0;
        Iterator<MyPos> tailToHead = segments.descendingIterator();
        while (tailToHead.hasNext()) {
            turnsUntilGone++;
            if (turnsUntilGrowth == 0) {
                turnsUntilGone++;
                turnsUntilGrowth = meta.getGrowFrequency();
            } else {
                turnsUntilGrowth--;
            }
            MyPos pos = tailToHead.next();
            // Det är ok att gå till pos så att du är där i turnsUntilGone drag
            g[pos.y][pos.x] = earlyTurn ? NewPaxport2.LETHAL : turnsUntilGone;
        }
        turnsUntilGrowthOnTurn = turnsUntilGone;
        turnsUntilGoneOnTurn = turnsUntilGone;
    }

    public void populateSnakeShrinkingWithExtraVag(int[][] g, HeatMeta meta, HeatState state, MyPos extraVag) {
        boolean earlyTurn = state.getTurn() <= 10;
        turnsUntilGone++;
        if (turnsUntilGrowth == 0) {
            turnsUntilGone++;
            turnsUntilGrowth = meta.getGrowFrequency();
        } else {
            turnsUntilGrowth--;
        }
        tmpOnExtraVag.add(g[extraVag.y][extraVag.x]);
        // Det är ok att gå till pos så att du är där i turnsUntilGone drag
        g[extraVag.y][extraVag.x] = earlyTurn ? NewPaxport2.LETHAL : turnsUntilGone;
    }

    public void clearExtraVag(int[][] g, List<MyPos> extraVag) {
        Iterator<Integer> extraIt = tmpOnExtraVag.iterator();
        for (MyPos pos : extraVag) {
            g[pos.y][pos.x] = extraIt.next();
        }
        tmpOnExtraVag.clear();
        turnsUntilGone = turnsUntilGrowthOnTurn;
        turnsUntilGone = turnsUntilGoneOnTurn;
    }
}
